;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "elfeed-org" "20250219.950"
  "Configure elfeed with one or more org-mode files."
  '((emacs  "28.1")
    (elfeed "1.1.1")
    (org    "8.2.7"))
  :url "https://github.com/remyhonig/elfeed-org"
  :commit "1197cf29f6604e572ec604874a8f50b58081176a"
  :revdesc "1197cf29f660"
  :keywords '("news")
  :authors '(("Remy Honig" . "remyhonig@gmail.com"))
  :maintainers '(("Remy Honig" . "remyhonig@gmail.com")))
